include machine.e
include file.e
include misc.e
include get.e
global constant DB_OK=0,
DB_OPEN_FAIL=-1,
DB_EXISTS_ALREADY=-2,
DB_LOCK_FAIL=-3
global constant DB_LOCK_NO=0,
DB_LOCK_SHARED=1,
DB_LOCK_EXCLUSIVE=2
constant DB_MAGIC=77
constant DB_MAJOR=0,DB_MINOR=5
constant SIZEOF_TABLE_HEADER=16
constant TABLE_HEADERS=3,FREE_COUNT=7,FREE_LIST=11
constant SCREEN=1
constant INIT_FREE=5,
INIT_TABLES=5,
INIT_INDEX=10,
INIT_RECORDS=20
constant TRUE=1
integer current_db
atom current_table
sequence db_names,db_file_nums,db_lock_methods
integer current_lock
current_db=-1
current_table=-1
db_names={}
db_file_nums={}
db_lock_methods={}
sequence key_pointers
procedure default_fatal(sequence msg)
puts(SCREEN,"Fatal Database Error: "&msg&'\n')
?1/0
end procedure
global integer db_fatal_id
db_fatal_id=routine_id("default_fatal")
procedure fatal(sequence msg)
call_proc(db_fatal_id,{msg})
end procedure
function get1()
return getc(current_db)
end function
function get4()
integer b0,b1,b2,b3
b0=getc(current_db)
b1=getc(current_db)
b2=getc(current_db)
b3=getc(current_db)
return b0+#100*b1+#10000*b2+#1000000*b3
end function
function get_string()
sequence s
integer c
s=""
while TRUE do
c=getc(current_db)
if c=-1 then
fatal("string is missing 0 terminator")
elsif c=0 then
exit
end if
s&=c
end while
return s
end function
constant I2B=249,
I3B=250,
I4B=251,
F4B=252,
F8B=253,
S1B=254,
S4B=255
constant MIN1B=-9,
MAX1B=239,
MIN2B=-power(2,15),
MAX2B=power(2,15)-1,
MIN3B=-power(2,23),
MAX3B=power(2,23)-1,
MIN4B=-power(2,31)
function decompress()
integer c
sequence s
atom len
c=getc(current_db)
if c>=0 and c<=248 then
return c+MIN1B
elsif c=I2B then
return getc(current_db)+
#100 * getc(current_db)+
MIN2B
elsif c=I3B then
return getc(current_db)+
#100 * getc(current_db)+
#10000 * getc(current_db)+
MIN3B
elsif c=I4B then
return get4()+MIN4B
elsif c=F4B then
return float32_to_atom({getc(current_db),getc(current_db),
getc(current_db),getc(current_db)})
elsif c=F8B then
return float64_to_atom({getc(current_db),getc(current_db),
getc(current_db),getc(current_db),
getc(current_db),getc(current_db),
getc(current_db),getc(current_db)})
else
if c=S1B then
len=getc(current_db)
else
len=get4()
end if
s=repeat(0,len)
for i=1 to len do
s[i]=decompress()
end for
return s
end if
end function
function compress(object x)
sequence x4,s
if atom(x)then
if integer(x)then
if x>=MIN1B and x<=MAX1B then
return{x-MIN1B}
elsif x>=MIN2B and x<=MAX2B then
x-=MIN2B
return{I2B,and_bits(x,#FF),floor(x/#100)}
elsif x>=MIN3B and x<=MAX3B then
x-=MIN3B
return{I3B,and_bits(x,#FF),and_bits(floor(x/#100),#FF),floor(x/#10000)}
else
return I4B&int_to_bytes(x-MIN4B)
end if
else
x4=atom_to_float32(x)
if x=float32_to_atom(x4)then
return F4B&x4
else
return F8B&atom_to_float64(x)
end if
end if
else
if length(x)<=255 then
s={S1B,length(x)}
else
s=S4B&int_to_bytes(length(x))
end if
for i=1 to length(x)do
s&=compress(x[i])
end for
return s
end if
end function
procedure put1(atom x)
puts(current_db,x)
end procedure
atom mem
mem=allocate(4)
procedure put4(atom x)
poke4(mem,x)
puts(current_db,peek({mem,4}))
end procedure
procedure putn(sequence s)
puts(current_db,s)
end procedure
procedure safe_seek(atom pos)
if seek(current_db,pos)!=0 then
fatal(sprintf("seek to position %d failed!\n",pos))
end if
end procedure
global procedure db_dump(integer fn,integer low_level_too)
integer magic,minor,major
atom tables,ntables,tname,trecords,t_header,tnrecs,
key_ptr,data_ptr,size,addr,tindex
object key,data
integer c,n,tblocks
atom a
if low_level_too then
safe_seek(0)
a=0
while TRUE do
c=getc(current_db)
if c=-1 then
exit
end if
if remainder(a,10)=0 then
printf(SCREEN,"\n%d: ",a)
end if
printf(SCREEN,"%d, ",c)
a+=1
end while
end if
puts(fn,'\n')
safe_seek(0)
magic=get1()
if magic!=DB_MAGIC then
puts(fn,"This is not a Euphoria Database file\n")
return
end if
major=get1()
minor=get1()
printf(fn,"Euphoria Database System Version %d.%d\n",{major,minor})
tables=get4()
safe_seek(tables)
ntables=get4()
printf(fn,"There are %d tables\n",ntables)
t_header=where(current_db)
for t=1 to ntables do
tname=get4()
tnrecs=get4()
tblocks=get4()
tindex=get4()
safe_seek(tname)
printf(fn,"%s:\n",{get_string()})
for b=1 to tblocks do
printf(fn,"block #%d\n",b)
safe_seek(tindex+(b-1)*8)
tnrecs=get4()
trecords=get4()
for r=1 to tnrecs do
safe_seek(trecords+(r-1)*4)
key_ptr=get4()
safe_seek(key_ptr)
data_ptr=get4()
key=decompress()
puts(fn,"  key: ")
print(fn,key)
puts(fn,' ')
safe_seek(data_ptr)
data=decompress()
puts(fn,"  data: ")
print(fn,data)
puts(fn,'\n')
end for
end for
t_header+=SIZEOF_TABLE_HEADER
safe_seek(t_header)
end for
puts(fn,"List of free blocks:\n")
safe_seek(FREE_COUNT)
n=get4()
safe_seek(get4())
for i=1 to n do
addr=get4()
size=get4()
printf(fn,"%d: %d bytes\n",{addr,size})
end for
end procedure
global procedure check_free_list()
atom free_count,free_list,addr,size,free_list_space
atom max
safe_seek(-1)
max=where(current_db)
safe_seek(FREE_COUNT)
free_count=get4()
if free_count>max/13 then
fatal("free count is too high")
end if
free_list=get4()
if free_list>max then
fatal("bad free list pointer")
end if
safe_seek(free_list-4)
free_list_space=get4()
if free_list_space>max or free_list_space<INIT_FREE * 8 then
fatal("free list space is bad")
end if
for i=1 to free_count do
safe_seek(free_list+(i-1)*8)
addr=get4()
if addr>max then
fatal("bad block address")
end if
size=get4()
if size>max then
fatal("block size too big")
end if
safe_seek(addr-4)
if get4()>size then
fatal("bad size in front of free block")
end if
end for
end procedure
function db_allocate(atom n)
atom free_list,size,size_ptr,addr
integer free_count
sequence remaining
safe_seek(FREE_COUNT)
free_count=get4()
if free_count>0 then
free_list=get4()
safe_seek(free_list)
for i=1 to free_count do
addr=get4()
size_ptr=where(current_db)
size=get4()
if size>=n+4 then
if size>=n+16 then
safe_seek(addr-4)
put4(size-n-4)
safe_seek(size_ptr)
put4(size-n-4)
addr+=size-n-4
safe_seek(addr-4)
put4(n+4)
else
remaining=get_bytes(current_db,(free_count-i)* 8)
safe_seek(free_list+8*(i-1))
putn(remaining)
safe_seek(FREE_COUNT)
put4(free_count-1)
safe_seek(addr-4)
put4(size)
end if
return addr
end if
end for
end if
safe_seek(-1)
put4(n+4)
return where(current_db)
end function
procedure db_free(atom p)
atom psize,i,size,addr,free_list,free_list_space
atom new_space,to_be_freed,prev_addr,prev_size
integer free_count
sequence remaining
safe_seek(p-4)
psize=get4()
safe_seek(FREE_COUNT)
free_count=get4()
free_list=get4()
safe_seek(free_list-4)
free_list_space=get4()-4
if free_list_space<8 *(free_count+1)then
new_space=floor(free_list_space * 3/2)
to_be_freed=free_list
free_list=db_allocate(new_space)
safe_seek(free_list-4)
safe_seek(FREE_COUNT)
free_count=get4()
safe_seek(FREE_LIST)
put4(free_list)
safe_seek(to_be_freed)
remaining=get_bytes(current_db,8*free_count)
safe_seek(free_list)
putn(remaining)
putn(repeat(0,new_space-length(remaining)))
safe_seek(free_list)
else
new_space=0
end if
i=1
prev_addr=0
prev_size=0
while i<=free_count do
addr=get4()
size=get4()
if p<addr then
exit
end if
prev_addr=addr
prev_size=size
i+=1
end while
if i>1 and prev_addr+prev_size=p then
safe_seek(free_list+(i-2)*8+4)
if i<free_count and p+psize=addr then
put4(prev_size+psize+size)
safe_seek(free_list+i*8)
remaining=get_bytes(current_db,(free_count-i)*8)
safe_seek(free_list+(i-1)*8)
putn(remaining)
free_count-=1
safe_seek(FREE_COUNT)
put4(free_count)
else
put4(prev_size+psize)
end if
elsif i<free_count and p+psize=addr then
safe_seek(free_list+(i-1)*8)
put4(p)
put4(psize+size)
else
safe_seek(free_list+(i-1)*8)
remaining=get_bytes(current_db,(free_count-i+1)*8)
free_count+=1
safe_seek(FREE_COUNT)
put4(free_count)
safe_seek(free_list+(i-1)*8)
put4(p)
put4(psize)
putn(remaining)
end if
if new_space then
db_free(to_be_freed)
end if
end procedure
global function db_create(sequence path,integer lock_method)
integer db
if not find('.',path)then
path&=".edb"
end if
db=open(path,"rb")
if db!=-1 then
close(db)
return DB_EXISTS_ALREADY
end if
db=open(path,"wb")
if db=-1 then
return DB_OPEN_FAIL
end if
close(db)
db=open(path,"ub")
if db=-1 then
return DB_OPEN_FAIL
end if
if lock_method=DB_LOCK_SHARED then
lock_method=DB_LOCK_NO
end if
if lock_method=DB_LOCK_EXCLUSIVE then
if not lock_file(db,LOCK_EXCLUSIVE,{})then
return DB_LOCK_FAIL
end if
end if
current_db=db
current_lock=lock_method
current_table=-1
db_names=append(db_names,path)
db_lock_methods=append(db_lock_methods,lock_method)
db_file_nums=append(db_file_nums,db)
put1(DB_MAGIC)
put1(DB_MAJOR)
put1(DB_MINOR)
put4(19)
put4(0)
put4(23+INIT_TABLES * SIZEOF_TABLE_HEADER+4)
put4(8+INIT_TABLES * SIZEOF_TABLE_HEADER)
put4(0)
putn(repeat(0,INIT_TABLES * SIZEOF_TABLE_HEADER))
put4(4+INIT_FREE*8)
putn(repeat(0,INIT_FREE * 8))
return DB_OK
end function
global function db_open(sequence path,integer lock_method)
integer db,magic
if not find('.',path)then
path&=".edb"
end if
if lock_method=DB_LOCK_NO or
lock_method=DB_LOCK_EXCLUSIVE then
db=open(path,"ub")
else
db=open(path,"rb")
end if
if db=-1 then
return DB_OPEN_FAIL
end if
if lock_method=DB_LOCK_EXCLUSIVE then
if not lock_file(db,LOCK_EXCLUSIVE,{})then
return DB_LOCK_FAIL
end if
elsif lock_method=DB_LOCK_SHARED then
if not lock_file(db,LOCK_SHARED,{})then
return DB_LOCK_FAIL
end if
end if
magic=getc(db)
if magic!=DB_MAGIC then
return DB_OPEN_FAIL
end if
current_db=db
current_table=-1
current_lock=lock_method
db_names=append(db_names,path)
db_lock_methods=append(db_lock_methods,lock_method)
db_file_nums=append(db_file_nums,db)
return DB_OK
end function
global function db_select(sequence path)
integer index
if not find('.',path)then
path&=".edb"
end if
index=find(path,db_names)
if index=0 then
return DB_OPEN_FAIL
end if
current_db=db_file_nums[index]
current_lock=db_lock_methods[index]
current_table=-1
return DB_OK
end function
global procedure db_close()
integer index
if current_lock then
unlock_file(current_db,{})
end if
close(current_db)
index=find(current_db,db_file_nums)
db_names=db_names[1..index-1]&
db_names[index+1..length(db_names)]
db_file_nums=db_file_nums[1..index-1]&
db_file_nums[index+1..length(db_file_nums)]
db_lock_methods=db_lock_methods[1..index-1]&
db_lock_methods[index+1..length(db_lock_methods)]
current_db=-1
end procedure
function table_find(sequence name)
atom tables,nt,t_header,name_ptr
sequence tname
safe_seek(TABLE_HEADERS)
tables=get4()
safe_seek(tables)
nt=get4()
t_header=tables+4
for i=1 to nt do
safe_seek(t_header)
name_ptr=get4()
safe_seek(name_ptr)
tname=get_string()
if equal(tname,name)then
return t_header
end if
t_header+=SIZEOF_TABLE_HEADER
end for
return-1
end function
global function db_select_table(sequence name)
atom table,nkeys,index
atom block_ptr,block_size
integer blocks
table=table_find(name)
if table=-1 then
return DB_OPEN_FAIL
end if
current_table=table
safe_seek(table+4)
nkeys=get4()
blocks=get4()
index=get4()
key_pointers={}
for b=0 to blocks-1 do
safe_seek(index+8 * b)
block_size=get4()
block_ptr=get4()
safe_seek(block_ptr)
for j=1 to block_size do
key_pointers&=get4()
end for
end for
return DB_OK
end function
global function db_create_table(sequence name)
atom name_ptr,nt,tables,newtables,table,records_ptr
atom size,newsize,index_ptr
sequence remaining
table=table_find(name)
if table!=-1 then
return DB_EXISTS_ALREADY
end if
safe_seek(TABLE_HEADERS)
tables=get4()
safe_seek(tables-4)
size=get4()
nt=get4()+1
if nt*SIZEOF_TABLE_HEADER+8>size then
newsize=floor(size * 3/2)
newtables=db_allocate(newsize)
put4(nt)
safe_seek(tables+4)
remaining=get_bytes(current_db,(nt-1)*SIZEOF_TABLE_HEADER)
safe_seek(newtables+4)
putn(remaining)
putn(repeat(0,newsize-4-(nt-1)*SIZEOF_TABLE_HEADER))
db_free(tables)
safe_seek(TABLE_HEADERS)
put4(newtables)
tables=newtables
else
safe_seek(tables)
put4(nt)
end if
records_ptr=db_allocate(INIT_RECORDS * 4)
putn(repeat(0,INIT_RECORDS * 4))
index_ptr=db_allocate(INIT_INDEX * 8)
put4(0)
put4(records_ptr)
putn(repeat(0,(INIT_INDEX-1)* 8))
name_ptr=db_allocate(length(name)+1)
putn(name&0)
safe_seek(tables+4+(nt-1)*SIZEOF_TABLE_HEADER)
put4(name_ptr)
put4(0)
put4(1)
put4(index_ptr)
if db_select_table(name)then
end if
return DB_OK
end function
global procedure db_delete_table(sequence name)
atom table,tables,nt,nrecs,records_ptr,blocks
atom p,data_ptr,index
sequence remaining
table=table_find(name)
if table=-1 then
return
end if
safe_seek(table)
db_free(get4())
safe_seek(table+4)
nrecs=get4()
blocks=get4()
index=get4()
for b=0 to blocks-1 do
safe_seek(index+b*8)
nrecs=get4()
records_ptr=get4()
for r=0 to nrecs-1 do
safe_seek(records_ptr+r*4)
p=get4()
safe_seek(p)
data_ptr=get4()
db_free(data_ptr)
db_free(p)
end for
db_free(records_ptr)
end for
db_free(index)
safe_seek(TABLE_HEADERS)
tables=get4()
safe_seek(tables)
nt=get4()
safe_seek(table+SIZEOF_TABLE_HEADER)
remaining=get_bytes(current_db,
tables+4+nt*SIZEOF_TABLE_HEADER-
(table+SIZEOF_TABLE_HEADER))
safe_seek(table)
putn(remaining)
nt-=1
safe_seek(tables)
put4(nt)
if table=current_table then
current_table=-1
elsif table<current_table then
current_table-=SIZEOF_TABLE_HEADER
end if
end procedure
global function db_table_list()
sequence table_names
atom tables,nt,name
safe_seek(TABLE_HEADERS)
tables=get4()
safe_seek(tables)
nt=get4()
table_names=repeat(0,nt)
for i=0 to nt-1 do
safe_seek(tables+4+i*SIZEOF_TABLE_HEADER)
name=get4()
safe_seek(name)
table_names[i+1]=get_string()
end for
return table_names
end function
function key_value(atom ptr)
safe_seek(ptr+4)
return decompress()
end function
global function db_find_key(object key)
atom lo,hi,mid,c
lo=1
hi=length(key_pointers)
mid=1
c=0
while lo<=hi do
mid=floor((lo+hi)/2)
c=compare(key,key_value(key_pointers[mid]))
if c<0 then
hi=mid-1
elsif c>0 then
lo=mid+1
else
return mid
end if
end while
if c>0 then
mid+=1
end if
return-mid
end function
global function db_insert(object key,object data)
sequence key_string,data_string,last_part,remaining
atom key_ptr,data_ptr,records_ptr,nrecs,current_block,size,new_size
atom key_location,new_block,index_ptr,new_index_ptr,total_recs
integer r,blocks,new_recs
key_location=db_find_key(key)
if key_location>0 then
return DB_EXISTS_ALREADY
end if
key_location=-key_location
data_string=compress(data)
key_string=compress(key)
data_ptr=db_allocate(length(data_string))
putn(data_string)
key_ptr=db_allocate(4+length(key_string))
put4(data_ptr)
putn(key_string)
if current_table=-1 then
fatal("no table selected")
end if
safe_seek(current_table+4)
total_recs=get4()+1
blocks=get4()
safe_seek(current_table+4)
put4(total_recs)
key_pointers=key_pointers[1..key_location-1]&
key_ptr&
key_pointers[key_location..length(key_pointers)]
safe_seek(current_table+12)
index_ptr=get4()
safe_seek(index_ptr)
r=0
while TRUE do
nrecs=get4()
records_ptr=get4()
r+=nrecs
if r+1>=key_location then
exit
end if
end while
current_block=where(current_db)-8
key_location-=(r-nrecs)
safe_seek(records_ptr+4*(key_location-1))
for i=key_location to nrecs+1 do
put4(key_pointers[i+r-nrecs])
end for
safe_seek(current_block)
nrecs+=1
put4(nrecs)
safe_seek(records_ptr-4)
size=get4()-4
if nrecs*4>size-4 then
new_size=8 * floor(sqrt(total_recs+10))
new_recs=floor(new_size/8)
if new_recs>floor(nrecs/2)then
new_recs=floor(nrecs/2)
end if
safe_seek(records_ptr+(nrecs-new_recs)*4)
last_part=get_bytes(current_db,new_recs*4)
new_block=db_allocate(new_size)
putn(last_part)
putn(repeat(0,new_size-length(last_part)))
safe_seek(current_block)
put4(nrecs-new_recs)
safe_seek(current_block+8)
remaining=get_bytes(current_db,index_ptr+blocks*8-(current_block+8))
safe_seek(current_block+8)
put4(new_recs)
put4(new_block)
putn(remaining)
safe_seek(current_table+8)
blocks+=1
put4(blocks)
safe_seek(index_ptr-4)
size=get4()-4
if blocks*8>size-8 then
remaining=get_bytes(current_db,blocks*8)
new_size=floor(size*3/2)
new_index_ptr=db_allocate(new_size)
putn(remaining)
putn(repeat(0,new_size-blocks*8))
db_free(index_ptr)
safe_seek(current_table+12)
put4(new_index_ptr)
end if
end if
return DB_OK
end function
global procedure db_delete_record(integer key_location)
atom key_ptr,nrecs,records_ptr,data_ptr,index_ptr,current_block
integer r,blocks
sequence remaining
if key_location<1 or key_location>length(key_pointers)then
fatal("bad record number")
end if
key_ptr=key_pointers[key_location]
safe_seek(key_ptr)
data_ptr=get4()
db_free(key_ptr)
db_free(data_ptr)
key_pointers=key_pointers[1..key_location-1]&
key_pointers[key_location+1..length(key_pointers)]
if current_table=-1 then
fatal("no table selected")
end if
safe_seek(current_table+4)
nrecs=get4()-1
blocks=get4()
safe_seek(current_table+4)
put4(nrecs)
safe_seek(current_table+12)
index_ptr=get4()
safe_seek(index_ptr)
r=0
while TRUE do
nrecs=get4()
records_ptr=get4()
r+=nrecs
if r>=key_location then
exit
end if
end while
r-=nrecs
current_block=where(current_db)-8
nrecs-=1
if nrecs=0 and blocks>1 then
remaining=get_bytes(current_db,index_ptr+blocks*8-(current_block+8))
safe_seek(current_block)
putn(remaining)
safe_seek(current_table+8)
put4(blocks-1)
db_free(records_ptr)
else
key_location-=r
safe_seek(current_block)
put4(nrecs)
safe_seek(records_ptr+4*(key_location-1))
for i=key_location to nrecs do
put4(key_pointers[i+r])
end for
end if
end procedure
global procedure db_replace_data(integer rn,object data)
atom old_size,new_size,key_ptr,data_ptr
sequence data_string
if rn<1 or rn>length(key_pointers)then
fatal("bad record number")
end if
key_ptr=key_pointers[rn]
safe_seek(key_ptr)
data_ptr=get4()
safe_seek(data_ptr-4)
old_size=get4()-4
data_string=compress(data)
new_size=length(data_string)
if new_size<=old_size and
new_size>=old_size-8 then
safe_seek(data_ptr)
else
db_free(data_ptr)
data_ptr=db_allocate(new_size)
safe_seek(key_ptr)
put4(data_ptr)
safe_seek(data_ptr)
end if
putn(data_string)
end procedure
global function db_table_size()
return length(key_pointers)
end function
global function db_record_data(integer rn)
atom data_ptr
object data_value
if rn<1 or rn>length(key_pointers)then
fatal("bad record number")
end if
safe_seek(key_pointers[rn])
data_ptr=get4()
safe_seek(data_ptr)
data_value=decompress()
return data_value
end function
global function db_record_key(integer rn)
atom data_ptr
object key_value
if rn<1 or rn>length(key_pointers)then
fatal("bad record number")
end if
safe_seek(key_pointers[rn])
data_ptr=get4()
key_value=decompress()
return key_value
end function
global function db_compress()
integer index,chunk_size,nrecs,r
sequence new_path,old_path,table_list,record,chunk
if current_db=-1 then
fatal("no current database")
end if
index=find(current_db,db_file_nums)
new_path=db_names[index]
old_path=new_path[1..length(new_path)-3]&"tmp"
db_close()
if platform()=LINUX then
system("mv "&new_path&" "&old_path,2)
else
system("ren "&new_path&" "&old_path,2)
end if
index=db_create(new_path,DB_LOCK_NO)
if index!=DB_OK then
if platform()=LINUX then
system("mv "&old_path&" "&new_path,2)
else
system("ren "&old_path&" "&new_path,2)
end if
return index
end if
index=db_open(old_path,DB_LOCK_NO)
table_list=db_table_list()
for i=1 to length(table_list)do
index=db_select(new_path)
index=db_create_table(table_list[i])
index=db_select(old_path)
index=db_select_table(table_list[i])
nrecs=db_table_size()
r=1
while r<=nrecs do
if nrecs-r+1>=20 then
chunk_size=20
else
chunk_size=nrecs-r+1
end if
chunk={}
for j=1 to chunk_size do
record={db_record_key(r),db_record_data(r)}
r+=1
chunk=append(chunk,record)
end for
index=db_select(new_path)
index=db_select_table(table_list[i])
for j=1 to chunk_size do
if db_insert(chunk[j][1],chunk[j][2])!=DB_OK then
fatal("couldn't insert into new database")
end if
end for
index=db_select(old_path)
index=db_select_table(table_list[i])
end while
end for
db_close()
index=db_select(new_path)
return DB_OK
end function
with 824413204
